# -*- coding: utf-8 -*-
"""
Contains functions not directly related to word parsing, but used by the it.
"""
from itertools import cycle

from .constants import PRASENS, PRATERITUM, PERSON_1, PERSON_2, PERSON_3


def recursively_extract(node, exfun, maxdepth=2):
    """
    Transform a html ul/ol tree into a python list tree.

    Converts a html node containing ordered and unordered lists and list items
    into an object of lists with tree-like structure. Leaves are retrieved by
    applying `exfun` function to the html nodes not containing any ul/ol list.

    Args:
        node: BeautifulSoup HTML node to traverse
        exfun: function to apply to every string node found
        maxdepth: maximal depth of lists to go in the node

    Returns:
        A tree-like python object composed of lists.


    Examples:

    >>> node_content = \
    '''
    <ol>
        <li>Hase</li>
        <li>Nase<ol><li>Eins</li><li>Zwei</li></ol></li>
    </ol>'''
    >>> node = BeautifulSoup(node_content, "lxml")
    >>> recursively_extract(node, lambda x: x)
    [<li>Hase</li>, [<li>Eins</li>, <li>Zwei</li>]]
    >>> recursively_extract(node, lambda x: x.get_text())
    ['Hase', ['Eins', 'Zwei']]
    """
    if node.name in ['ol', 'ul']:
        lilist = node
    else:
        lilist = node.ol or node.ul
    if lilist and maxdepth:
        # apply 'recursively_extract' to every 'li' node found under this node
        return [recursively_extract(li, exfun, maxdepth=(maxdepth - 1))
                for li in lilist.find_all('li', recursive=False)]
    # if this node doesn't contain 'ol' or 'ul' node, return the transformed
    # leaf (using the 'exfun' function)
    return exfun(node)


def clear_text(text):
    """
    Remove soft hyphens anywhere, and heading and trailing spaces.
    """
    return text.replace('\xad', '').strip()


def table_node_to_tagged_cells(table_node):
    """
    Takes a table HTML node and returns the list of table cell strings
    tagged using the table top and left header (optionally using the table
    name found in the upper-leftmost cell).

    The return type is a list of 2-tuples:
    [(tag_set, text), ...]

    where text is a string taken from the cell, and tag_set is a set of
    strings (tags). If e.g. cell in the 3rd row and 2nd column with the
    text 'der Barmherzigkeit', has its top_header tag (1st row, 2nd
    column) 'Singular' and its left_header tag (1st column, 3rd row)
    'Genitiv', the corresponding tuple would look like:
    ({'Singular', 'Genitiv'}, 'der Barmherzigkeit')
    .

    The first row is considered a header row, if it's inside of <thead>
    html tag. The first column is considered a header column if the
    corresponding cells are <th> html nodes.
    """
    # pylint: disable=too-many-locals
    left_header = []
    top_header = None
    table_content = []

    title_element = table_node.h3
    table_name = title_element.text if title_element else ''

    # convert table html node to raw table (list of lists) and optional
    # left and top headers (also lists)
    if table_node.thead:
        all_ths = table_node.thead.find_all(
            'th', class_='wrap-table__flexions-head')
        top_header = [clear_text(t.text) for t in all_ths]

    for row in table_node.tbody.find_all('tr'):
        if row.th:
            th_node = row.th
            rowspan = int(th_node.attrs.get('rowspan', 1))
            left_header.extend([clear_text(row.th.text)] * rowspan)

        tds = row.find_all('td')
        table_content.append([clear_text(td.text) for td in tds])

    # convert left, top, and table headers to sets for easier tagging
    if left_header:
        left_header = [{cell} for cell in left_header]
    else:
        left_header = [set() for _ in table_content]
    if top_header:
        top_header = [{cell} for cell in top_header]
    else:
        top_header = [set() for _ in table_content[0]]
    table_tag = {table_name} if table_name else set()

    if table_name in [PRASENS, PRATERITUM]:
        person_tags = [{PERSON_1}, {PERSON_2}, {PERSON_3}]
    else:
        person_tags = [set(), set(), set()]

    # create a list of tagged strings
    tagged_strings = []
    for row, row_tag, person_tag \
            in zip(table_content, left_header, cycle(person_tags)):
        for cell, col_tag in zip(row, top_header):
            taglist = table_tag \
                .union(row_tag) \
                .union(col_tag) \
                .union(person_tag)
            tagged_strings.append((taglist, cell))
    return tagged_strings



# -*- coding: utf-8 -*-
"""
CLI related functions
"""

import argparse
import sys

from crayons import white, blue, red  # pylint: disable=no-name-in-module
import yaml

from .__version__ import __version__
from .search import get, search
from .display import (display_grammar, display_compounds, print_tree_of_strings,
                      print_string_or_list, describe_word)


def display_word(word, args):
    """
    Display word attribute or general description, based on commandline arguments
    """
    # pylint: disable=too-many-branches
    if args.title:
        print(word.title)
    elif args.name:
        print(word.name)
    elif args.article:
        if word.article:
            print(word.article)
    elif args.part_of_speech:
        if word.part_of_speech:
            print(word.part_of_speech)
    elif args.frequency:
        if word.frequency:
            print(word.frequency)
    elif args.usage:
        if word.usage:
            print(word.usage)
    elif args.word_separation:
        for part in word.word_separation:
            print(part)
    elif args.meaning_overview:
        if word.meaning_overview:
            print_tree_of_strings(word.meaning_overview)
    elif args.synonyms:
        synonyms = word.synonyms
        if synonyms:
            print_string_or_list(synonyms)
    elif args.origin:
        if word.origin:
            print(word.origin)
    elif args.compounds:
        if word.compounds:
            display_compounds(word, args.compounds)
    elif args.phonetic:
        if word.phonetic:
            print(word.phonetic)
    elif args.alternative_spellings:
        if word.alternative_spellings:
            for spelling in word.alternative_spellings:
                print(spelling)
    elif args.grammar:
        display_grammar(word, args.grammar)
    elif args.export:
        yaml_string = yaml.dump(word.export(),
                                sort_keys=False, allow_unicode=True)
        print(yaml_string, end='')
    elif args.words_before:
        print_string_or_list(word.words_before)
    elif args.words_after:
        print_string_or_list(word.words_after)
    else:
        # print the description
        describe_word(word)


def parse_args():
    """
    Parse CLI arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('word')
    parser.add_argument('--title', action='store_true',
                        help=_('display word and article'))
    parser.add_argument('--name', action='store_true',
                        help=_('display the word itself'))
    parser.add_argument('--article', action='store_true',
                        help=_('display article'))
    parser.add_argument('--part-of-speech', action='store_true',
                        help=_('display part of speech'))
    parser.add_argument('--frequency', action='store_true',
                        help=_('display commonness (1 to 5)'))
    parser.add_argument('--usage', action='store_true',
                        help=_('display context of use'))
    parser.add_argument('--word-separation', action='store_true',
                        help=_('display proper separation (line separated)'))
    parser.add_argument('--meaning-overview', action='store_true',
                        help=_('display meaning overview'))
    parser.add_argument('--synonyms', action='store_true',
                        help=_('list synonyms (line separated)'))
    parser.add_argument('--origin', action='store_true',
                        help=_('display origin'))
    parser.add_argument('--compounds', nargs='?', const='ALL',
                        help=_('list common compounds'))
    parser.add_argument('-g', '--grammar', nargs='?', const='ALL',
                        help=_('list grammar forms'))
    parser.add_argument('--export', action='store_true',
                        help=_('export parsed word attributes in yaml format'))
    parser.add_argument('--words-before', action='store_true',
                        help=_('list 5 words before this one'))
    parser.add_argument('--words-after', action='store_true',
                        help=_('list 5 words after this one'))

    parser.add_argument('-r', '--result', type=int,
                        help=_('display n-th (starting from 1) result in case '
                               'of multiple words matching the input'))
    parser.add_argument('--fuzzy', action='store_true',
                        help=_('enable fuzzy word matching'))
    parser.add_argument('--no-cache', action='store_false', dest='cache',
                        help=_('do not cache retrieved words'))

    parser.add_argument('-V', '--version', action='store_true',
                        help=_('print program version'))
    parser.add_argument('--phonetic', action='store_true',
                        help=_('display pronunciation'))
    parser.add_argument('--alternative-spellings', action='store_true',
                        help=_('display alternative spellings'))

    return parser.parse_args()


def main():
    """
    Take the first CLI argument and describe the corresponding word
    """

    # handle the --version switch
    if '--version' in sys.argv or '-V' in sys.argv:
        print('duden ' + __version__)
        sys.exit(0)

    # parse normal arguments
    args = parse_args()

    # search all words matching the string
    words = search(args.word, return_words=False, exact=not args.fuzzy,
                   cache=args.cache)

    # exit if the word wasn't found
    if not words:
        print(red(_("Word '{}' not found")).format(args.word))
        sys.exit(1)

    # list the options when there is more than one matching word
    if len(words) > 1 and args.result is None:
        print(_('Found {} matching words. Use the -r/--result argument to '
                'specify which one to display.').format(white(len(words),
                                                              bold=True)))
        for i, word in enumerate(words, 1):
            print('{} {}'.format(blue('{})'.format(i)), word))
        sys.exit(1)

    result_index = args.result if args.result is not None else 1

    # choose the correct result
    try:
        word_url_suffix = words[result_index - 1]
    except IndexError:
        print(red(_("No result with number {}.")).format(result_index))
        sys.exit(1)

    # fetch and parse the word
    try:
        word = get(word_url_suffix, cache=args.cache)
    except Exception as exception:  # pylint: disable=broad-except
        print(red(exception))
        sys.exit(1)

    display_word(word, args)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
The duden package can parse the https://www.duden.de/ word information.

The `get` function is used to return parsed word, when provided with the word's
exact url name. The `search` function is used to search for words, either
returning exact matches (homonyms), or if fuzzy search is enabled, similar
words.

The basic class representing the parsed word is `DudenWord`.
"""
__all__ = [
    'get', 'search', 'get_word_of_the_day',
    'SINGULAR', 'PLURAL',
    'PRASENS', 'PRATERITUM',
    'INDIKATIV', 'IMPERATIV', 'KONJUKTIV_1', 'KONJUKTIV_2',
    'PARTIZIP_1', 'PARTIZIP_2', 'INFINITIV_MIT_ZU',
    'PERSON_1', 'PERSON_2', 'PERSON_3',
    'NOMINATIV', 'GENITIV', 'DATIV', 'AKKUSATIV'
]

from .search import get, search, get_word_of_the_day

from .constants import (SINGULAR, PLURAL,
                        PRASENS, PRATERITUM,
                        INDIKATIV, IMPERATIV, KONJUKTIV_1, KONJUKTIV_2,
                        PARTIZIP_1, PARTIZIP_2, INFINITIV_MIT_ZU,
                        PERSON_1, PERSON_2, PERSON_3,
                        NOMINATIV, GENITIV, DATIV, AKKUSATIV)



# -*- coding: utf-8 -*-
"""
Defines constants
"""

# grammar forms constants
SINGULAR = 'Singular'
PLURAL = 'Plural'

PRASENS = 'Präsens'
PRATERITUM = 'Präteritum'

INDIKATIV = 'Indikativ'
IMPERATIV = 'Imperativ'
KONJUKTIV_1 = 'Konjunktiv I'
KONJUKTIV_2 = 'Konjunktiv II'

PARTIZIP_1 = 'Partizip I'
PARTIZIP_2 = 'Partizip II'
INFINITIV_MIT_ZU = 'Infinitiv mit zu'

PERSON_1 = 'Person I'
PERSON_2 = 'Person II'
PERSON_3 = 'Person III'

NOMINATIV = 'Nominativ'
GENITIV = 'Genitiv'
DATIV = 'Dativ'
AKKUSATIV = 'Akkusativ'



# -*- coding: utf-8 -*-
"""
Console printing functions
"""
from string import ascii_lowercase

from crayons import white, blue, yellow  # pylint: disable=no-name-in-module


def display_grammar(word, grammar_args):
    """
    Display word grammar forms, corresponds to --grammar switch
    """
    grammar_struct = word.grammar_raw
    if not grammar_struct:
        return

    grammar_tokens = [token.lower() for token in grammar_args.split(',')]

    # filter out grammar forms which do not match provided keys
    tag_columns = []
    value_column = []
    for keys, value in word.grammar_raw:
        lkeys = {key.lower() for key in keys}

        if not (grammar_args == 'ALL' or lkeys.issuperset(grammar_tokens)):
            continue

        reduced_keys = lkeys.difference(grammar_tokens)

        tag_columns.append(list(reduced_keys))
        value_column.append(value)

    # determine the width of the table
    max_keys_count = max(map(len, tag_columns))

    # if provided keys uniquely determine the value(s), display a 1-col table
    if max_keys_count == 0:
        display_table([[value] for value in value_column])
        return

    # otherwise make a nice "| key1 key2 | value |" table
    table = []
    for keys, value in zip(tag_columns, value_column):
        padding = [""] * (max_keys_count - len(keys))
        row = keys + padding + [blue("|")] + [value]
        table.append(row)

    display_table(table)


def display_table(table, cell_spacing=' '):
    """
    Display general table with aligned columns
    """
    cols = list(zip(*table))
    collens = [max(len(word) for word in col) for col in cols]

    for row in table:
        for elem, collen in zip(row, collens):
            print(elem.ljust(collen), end=cell_spacing)
        print()


def display_compounds(word, compound_type='ALL'):
    """
    Print word common compounds
    """
    if compound_type == 'ALL':
        for part_of_speech, compounds in word.compounds.items():
            print(white('# ' + part_of_speech.capitalize(), bold=True))
            print_string_or_list(compounds)
            print()
    else:
        print_string_or_list(word.compounds[compound_type])


def print_tree_of_strings(tree):
    """
    Print a tree of strings up to depth 2

    Args:
        tree: tree of strings

    Example:

    >>> print_tree_of_strings(['Hase', ['Eins', 'Zwei']])
    0. Hase
    <BLANKLINE>
    1.  a. Eins
        b. Zwei
    """
    if isinstance(tree, str):
        print(tree)
        return

    for index, value in enumerate(tree):
        if isinstance(value, str):
            print(blue("{:>2}. ".format(index)), value, sep='')
        elif isinstance(value, list):
            for index_inner, value_inner in zip(ascii_lowercase, value):
                indent = blue("{:>2}. ".format(index)) if index_inner == 'a' else " " * 4
                print("{} {}".format(indent, blue(index_inner)), blue('. '), value_inner, sep='')
        print()


def print_string_or_list(obj):
    """
    Print string value or all list elements, if of type list
    """
    if isinstance(obj, list):
        for elem in obj:
            print(elem)
    else:
        print(obj)


def describe_word(word):
    """
    Print overall word description
    """
    print(yellow(word.title, bold=True))
    print(yellow('=' * len(word.title)))

    if word.phonetic:
        print(white(_('Pronunciation:'), bold=True), word.phonetic)
    if word.alternative_spellings:
        print(white(_('Alternative spellings:'), bold=True), ', '.join(word.alternative_spellings))
    if word.part_of_speech:
        print(white(_('Word type:'), bold=True), word.part_of_speech)
    if word.usage:
        print(white(_('Usage:'), bold=True), word.usage)
    if word.frequency:
        commonness = '{label} {frequency}{max_frequency}'.format(
            label=white(_('Commonness:'), bold=True),
            frequency=word.frequency,
            max_frequency=blue('/5'))
        print(commonness)
    if word.word_separation:
        print('{label} {content}'.format(
            label=white(_('Separation:'), bold=True),
            content=str(blue('|')).join(word.word_separation)))

    if word.meaning_overview:
        print(white(_('Meaning overview:'), bold=True))
        print_tree_of_strings(word.meaning_overview)

    if word.synonyms:
        print(white(_('Synonyms:'), bold=True))
        print_tree_of_strings(word.synonyms)

    if word.compounds:
        print(white(_('Typical compounds:'), bold=True))
        for part_of_speech, words in word.compounds.items():
            print(blue(' - {}:'.format(part_of_speech.capitalize())),
                  ', '.join(words))



# -*- coding: utf-8 -*-
from pathlib import Path
from setuptools import setup

this_directory = Path(__file__).parent

about = {}
version_file = this_directory / "duden" / "__version__.py"
with open(version_file, 'r', encoding='utf-8') as f:
    exec(f.read(), about)

long_description = (this_directory / "README.md").read_text()

setup(
    author='Radomír Bosák',
    author_email='radomir.bosak@gmail.com',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Natural Language :: German',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Education',
    ],
    description='CLI-based german dictionary',
    long_description=long_description,
    long_description_content_type='text/markdown',
    download_url='https://github.com/radomirbosak/duden/archive/' \
                 + about['__version__'] + '.tar.gz',
    entry_points={
        'console_scripts': [
            'duden = duden.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=[
        "requests",
        "beautifulsoup4",
        "crayons",
        "pyxdg",
        "pyyaml"
    ],
    keywords=['duden', 'duden.de', 'dictionary', 'cli', 'word', 'german'],
    license='MIT',
    name='duden',
    packages=['duden'],
    url='https://github.com/radomirbosak/duden',
    version=about['__version__'],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Contains the DudenWord class: a parser of duden.de response.
"""

import copy
import gettext
import os

from .common import recursively_extract, table_node_to_tagged_cells, clear_text


EXPORT_ATTRIBUTES = [
    'name', 'urlname', 'title', 'article', 'part_of_speech', 'usage',
    'frequency', 'word_separation', 'meaning_overview', 'origin', 'compounds',
    'grammar_raw', 'synonyms', 'words_before', 'words_after', 'phonetic', 'alternative_spellings'
]

gettext.install('duden', os.path.join(os.path.dirname(__file__), 'locale'))


class DudenWord():
    """
    Represents parsed word. Takes a BeautifulSoup object as a constructor argument.

    Example:

        > r = requests.get('https://www.duden.de/rechtschreibung/Hase')
        > soup = bs4.BeautifulSoup(r.text)
        > word = duden.DudenWord(soup)
        > word
        Hase, der (Substantiv, maskulin)
    """
    # pylint: disable=too-many-public-methods
    wordcloud_parts_of_speech = ['substantive', 'verben', 'adjektive']

    def __init__(self, soup):
        self.soup = soup

    def __repr__(self):
        return '{} ({})'.format(self.title, self.part_of_speech)

    @property
    def title(self):
        """
        The word string with article
        """
        return self.soup.h1.get_text().replace('\xad', '').strip()

    @property
    def name(self):
        """
        Word without article
        """

        # Find span with class "lemma__main"
        title_element = self.soup.find('span', {"class": "lemma__main"})
        if title_element is not None:
            # remove soft hyphens "\xad" and return
            return clear_text(title_element.get_text())

        #  if the title_element does not exist, we fall back to the old method
        if self.part_of_speech is not None and 'Substantiv' not in self.part_of_speech:
            return self.title
        if ', ' not in self.title:
            return self.title

        name, _ = self.title.split(', ', 1)
        return name

    @property
    def urlname(self):
        """
        Return unique representation of the word used in duden.de urls
        """
        return self.soup.head.find('link', rel='canonical').attrs['href'].split('/')[-1]

    @property
    def revision_url(self):
        """Returns url to this specific word revision"""
        return self.soup.find('input', id='cite-field').attrs['value']

    @property
    def node_no(self):
        """Returns word node number"""
        return self.revision_url.split('/')[-3]

    @property
    def revision_no(self):
        """Returns word revision number"""
        return self.revision_url.split('/')[-1]

    @property
    def article(self):
        """
        Word article
        """
        # Find span with class "lemma__determiner"
        article_element = self.soup.find('span', {"class": "lemma__determiner"})
        if article_element is not None:
            # remove soft hyphens "\xad" and return
            return clear_text(article_element.get_text())

        #  if the article_element does not exist, we fall back to the old method
        if self.part_of_speech is not None and 'Substantiv' not in self.part_of_speech:
            return None
        if ', ' not in self.title:
            return None

        _, article = self.title.split(', ', 1)
        return article

    def _find_tuple_dl(self, key, element=None):
        """
        Get value element corresponding to key element containing the text
        provided by the `key` argument
        """
        if element is None:
            element = self.soup.article

        dls = element.find_all('dl', class_='tuple', recursive=False)
        for dl_node in dls:
            label = dl_node.find('dt', class_='tuple__key')
            if key in label.text:
                return dl_node.find('dd', class_='tuple__val')

        return None

    @property
    def part_of_speech(self):
        """
        Return the part of speech
        """
        try:
            pos_element = self._find_tuple_dl('Wortart')
            return pos_element.text
        except AttributeError:
            return None

    @property
    def frequency(self):
        """
        Return word frequency:

        0 - least frequent
        5 - most frequent
        """
        try:
            freq_bar = self.soup.find('span', class_='shaft__full')
            return len(freq_bar.text)
        except AttributeError:
            return None

    @property
    def usage(self):
        """
        Return usage context
        """
        try:
            element = self._find_tuple_dl('Gebrauch')
            return element.text
        except AttributeError:
            return None

    @property
    def word_separation(self):
        """
        Return the word separated in a form of a list
        """
        containing_div = self.soup.find('div', id='rechtschreibung')
        sep_element = self._find_tuple_dl('Worttrennung', containing_div)
        if not sep_element:
            return None

        return sep_element.text.split('|')

    @property
    def meaning_overview(self):
        """
        Return the meaning structure, which can be string, list or a dict
        """
        section = self.soup.find('div', id='bedeutung') \
            or self.soup.find('div', id='bedeutungen')
        if section is None:
            return None
        section = copy.copy(section)
        section.header.extract()

        # 1. remove examples
        for dl_node in section.find_all('dl', class_='note'):
            # pylint: disable=condition-evals-to-constant
            if True or dl_node.dt.text == 'Beispiele':
                dl_node.extract()

        # 2. remove grammar parts
        for dl_node in section.find_all('dl', class_='tuple'):
            if dl_node.dt.text in ['Grammatik', 'Gebrauch']:
                dl_node.extract()

        # 3. remove pictures
        for node in section.find_all('figure'):
            node.extract()

        return recursively_extract(
            section, maxdepth=2, exfun=lambda x: x.text.strip())

    @property
    def synonyms(self):
        """
        Return the structure with word synonyms
        """
        try:
            section = self.soup.find('div', id='synonyme')
            section = copy.copy(section)
            if section.header:
                section.header.extract()
            return recursively_extract(section, maxdepth=2,
                                       exfun=lambda x: x.text.strip())
        except AttributeError:
            return None

    @property
    def origin(self):
        """
        Return the word origin
        """
        section = self.soup.find('div', id='herkunft')
        if section is None:
            return None

        section = copy.copy(section)
        if section.header:
            section.header.extract()
        return section.text.strip()

    @property
    def compounds(self):
        """
        Return the typical word compounds
        """
        section = self.soup.find('div', id='kontext')
        if not section:
            return None

        pos_trans = {
            'noun': 'substantive',
            'verb': 'verben',
            'adj': 'adjektive',
        }

        compounds = {}

        cluster_element = section.find('figure', class_='tag-cluster__cluster')
        for a_node in cluster_element.find_all('a'):
            compound_word = a_node.text
            compound_type = pos_trans[a_node.attrs['data-group']]

            if compound_type not in compounds:
                compounds[compound_type] = []

            compounds[compound_type].append(compound_word)

        compounds_sorted = {}
        for pos in sorted(compounds.keys()):
            compounds_sorted[pos] = sorted(compounds[pos])

        return compounds_sorted

    def grammar(self, *target_tags):
        """
        Return the information from grammar section

        Example:
        >>> word_laufen.grammar(duden.SINGULAR, duden.PRASENS, \
                                duden.INDIKATIV, duden.PERSON_3)
        ['er/sie/es läuft']
        """
        tagged_strings = self.grammar_raw
        target_tags = set(target_tags)
        return [string
                for tags, string in tagged_strings
                if target_tags.issubset(tags)]

    @property
    def grammar_raw(self):
        """
        Find the Grammar sections in the document and extract tagged string
        list of all tables found there.

        The concatenated tagged string list (for all tables) is returned
        """
        section = self.soup.find('div', id='grammatik')
        if not section:
            return []

        table_nodes = self.soup.find_all('div', class_='wrap-table') \
            + self.soup.find_all('table', class_='mere-table')

        tagged_strings = []
        for table_node in table_nodes:
            tagged_strings.extend(table_node_to_tagged_cells(table_node))
        return tagged_strings

    def export(self):
        """
        Export word's attributes as a dictionary

        Used e.g. for creating test data.
        """
        worddict = {}
        for attribute in EXPORT_ATTRIBUTES:
            worddict[attribute] = getattr(self, attribute, None)

        # convert grammar to lists
        if worddict['grammar_raw'] is not None:
            listed_grammar = []
            for keylist, form in worddict['grammar_raw']:
                listed_grammar.append([sorted(keylist), form])
            worddict['grammar_raw'] = listed_grammar
        return worddict

    @property
    def before_after_structure(self):
        """
        Parsed "Blätter section"

        Returns: dict mapping section names to list of words tuples.
                 Each tuple is comprised of word name and word urlname.

        Example:
            >>> duden.get("laufen").before_after_structure
            {'Im Alphabet davor': [('Laufbekleidung', 'Laufbekleidung'),
              ('Laufbrett', 'Laufbrett'),
              ('Laufbursche', 'Laufbursche'),
              ('Läufchen', 'Laeufchen'),
              ('Läufel', 'Laeufel')],
             'Im Alphabet danach': [('laufend', 'laufend'),
              ('laufen lassen, laufenlassen', 'laufen_lassen'),
              ('Laufer', 'Laufer'),
              ('Läufer', 'Laeufer'),
              ('Lauferei', 'Lauferei')]}
        """
        result = {}
        section = self.soup.find('div', id='block-beforeafterblock-2')
        for group in section.find_all('nav', class_='hookup__group'):
            h3title = group.h3.text
            result[h3title] = []
            for item in group.find_all('li'):
                link = item.a.attrs['href'].split('/')[-1]
                result[h3title].append((clear_text(item.text), link))
        return result

    @property
    def words_before(self):
        """Returns 5 words before this one in duden database"""
        return [name for name, _ in self.before_after_structure['Im Alphabet davor']]

    @property
    def words_after(self):
        """Returns 5 words after this one in duden database"""
        return [name for name, _ in self.before_after_structure['Im Alphabet danach']]

    @property
    def phonetic(self):
        """
        Returns pronunciation of the word in phonetic notation.
        See: https://en.wikipedia.org/wiki/International_Phonetic_Alphabet
        """
        ipa = self.soup.find('span', {"class": "ipa"})
        if ipa is not None:
            return ipa.get_text()

        return None

    @property
    def alternative_spellings(self):
        """
        Returns alternate spellings
        """
        alternative_spellings = self.soup.find_all('span', {"class": "lemma__alt-spelling"})
        if alternative_spellings is None:
            return None

        return [spelling.get_text() for spelling in alternative_spellings]
